#include <stdio.h>

  /* Parcijalne sume harmonijskog reda = zbroj prvih n clanova.
     Usporedba raznih algoritama (zbrajanje unaprijed i unatrag)
     i raznih tipova realne aritmetike (double, float). */

  /* Napomena: program je NAMJERNO napisan tako da se SVI medjurezultati
     (poput 1.0 / i) spremaju u varijable odgovarajuceg tipa, zato da se
     POKUSA dobiti konzistentno zaokruzivanje u pripadnom tipu, prilikom
     cijelog racuna.
       Nazalost, BEZ obzira na to, rezultati BITNO ovise o OPCIJAMA
     koje su zadane prilikom kompajliranja (naravno, i o kompajleru).
       Dobiveni rezultati mogu biti TOCNIJI nego sto bi trebali biti,
     jer se stvarno racunanje i zbrajanje clanova moze raditi u registrima
     bitno VECE tocnosti nego sto je zadani tip, a tek FINALNI rezultat
     se zaokruzuje na zadani tip (double, float). */


/****************************************************************************/

    /* Referentne vrijednosti u najpreciznijem tipu = double. */


double Priest_DC_Sum(int n)
{
    /* Priest's doubly compensated summation (DCS):

       The order of terms must be decreasing in absolute value,
       abs( x_1 ) >= abs( x_2 ) >= ... >= abs( x_n ).

           Sum := 1 / 1 +  ... + 1 / n  (decreasing absolute values).

       Error bound:

           abs( Sum - fl( Sum ) ) <= 2 * u * abs( Sum ).

       Algorithm 6.8, p. 194, in:
         J.-M. Muller, N. Brisebarre, F. de Dinechin, C.-P. Jeannerod,
           V. Lef\`evre, G. Melquiond, N. Revol, D. Stehl\'e and S. Torres:
         ``Handbook of Floating-Point Arithmetic'',
         Birkh\"auser, Boston, 2010.

       Algorithm 4.3, pp. 87--88, in:
         N. J. Higham:
         ``Accuracy and Stability of Numerical Algorithms, Second Edition'',
         SIAM, Philadelphia, USA, 2002.
    */

    int i;
    double Clan;
    double s, c, t, u, v, y, z;

    s = 1.0;
    c = 0.0;
    for (i = 2; i <= n; ++i) {

        Clan = 1.0 / ( (double) i );    /* Clan x_i u tipu double. */

        y = c + Clan;
        u = Clan - ( y - c );
        t = y + s ;
        v = y - ( t - s );
        z = u + v;
        s = t + z;
        c = z - ( s - t );
    }

    return s;
}

/****************************************************************************/

double Rump_Ogita_Oishi_Sum(int n)
{
    /* Pichat and Neumaier's summation, or
       Rump, Ogita, Oishi cascaded summation (ROO):

       The terms can be in any order - we use the natural (decreasing) order.

           Sum := 1 / 1 +  ... + 1 / n  (decreasing values).

       Error bound:

           abs( Sum - fl( Sum ) ) <= u * Sum
                                      + sqr( gamma_(n - 1) ) * Sum_abs_values.

       Here: Sum = abs( Sum ) = Sum_abs_values.

       Algorithms 6.10, 6.11, pp. 195--196, in:
         J.-M. Muller, N. Brisebarre, F. de Dinechin, C.-P. Jeannerod,
           V. Lef\`evre, G. Melquiond, N. Revol, D. Stehl\'e and S. Torres:
         ``Handbook of Floating-Point Arithmetic'',
         Birkh\"auser, Boston, 2010.
    */

    int i;
    double Clan;
    double s, e, ei, a, b, a1, b1, da, db;

    s = 1.0;
    e = 0.0;
    for (i = 2; i <= n; ++i) {

        Clan = 1.0 / ( (double) i );    /* Clan x_i u tipu double. */

            /* (s, ei) = 2Sum(s, x_i); */
        a = s;
        b = Clan;

        s = a + b;
        a1 = s - b;
        b1 = s - a1;
        da = a - a1;
        db = b - b1;
        ei = da + db;

        e = e + ei;
    }

    return s + e;
}

/****************************************************************************/

int main(void)
{
    int n = 1000000;    /* Broj clanova reda u sumi. */
    int i;

    double Sum, Sum_ROO;
    double err_1, err_2, rel_err_1, rel_err_2;

    double clan_d, sum_1_d, sum_2_d;
    float  clan_f, sum_1_f, sum_2_f;

    double e_d, t_d, y_d, c_sum_1_d, c_sum_2_d;
    float  e_f, t_f, y_f, c_sum_1_f, c_sum_2_f;

    double e_1_d, e_2_d, k_sum_1_d, k_sum_2_d;
    float  e_1_f, e_2_f, k_sum_1_f, k_sum_2_f;

    FILE* out;

/****************************************************************************/

    /* Vrijednosti u raznim tipovima. */


    /* Algorithms:
         - plain iterative (recursive) summation,
         - Kahan's compensated summation without and with correction
           (see description below).
       Tested in two orders of terms in the sum:
         - forward = decreasing order (less stable for iterative algorithm),
         - backward = decreasing order (more stable for iterative algorithm).
    */


    /* Kahan's compensated summation (without and with correction):

       The terms can be in any order. We use decreasing and increasing order.

           Sum := 1 / 1 +  ... + 1 / n  (decreasing values),
           Sum := 1 / n +  ... + 1 / 1  (increasing values).

       Error bound:

           abs( Sum - fl( Sum ) ) <= ( 2 * u + O(nu^2) ) * Sum_abs_values.

       Here: Sum = abs( Sum ) = Sum_abs_values.

       Algorithms 6.6, 6.7, pp. 192--193, in:
         J.-M. Muller, N. Brisebarre, F. de Dinechin, C.-P. Jeannerod,
           V. Lef\`evre, G. Melquiond, N. Revol, D. Stehl\'e and S. Torres:
         ``Handbook of Floating-Point Arithmetic'',
         Birkh\"auser, Boston, 2010.

       Algorithm 4.2 + Kahan's final correction, pp. 84--85:
         N. J. Higham:
         ``Accuracy and Stability of Numerical Algorithms, Second Edition'',
         SIAM, Philadelphia, USA, 2002.
    */


    out = fopen("h_sum_alg.out", "w");

    fprintf(out, "\n");
    fprintf(out, " Suma prvih n clanova harmonijskog reda\n");
    fprintf(out, " zbrajanjem unaprijed (->) i unatrag (<-)\n");
    fprintf(out, "\n");
    fprintf(out, " Broj clanova reda  n = %d\n", n );
    fprintf(out, "\n");


        /***** Referentne vrijednosti - double (najprecizniji tip). *****/

        /* Double --- Priest's doubly compensated summation. */

    Sum = Priest_DC_Sum(n);

    fprintf(out, "\n");
    fprintf(out, " Double (maksimalno tocno):\n");
    fprintf(out, "   DCS (->) = %18.15f\n", Sum);


        /* Double --- Cascaded summation. */

    Sum_ROO = Rump_Ogita_Oishi_Sum(n);

    err_1 = Sum - Sum_ROO;
    rel_err_1 = err_1 / Sum;

    fprintf(out, "\n");
    fprintf(out, " Double (kaskadno zbrajanje):\n");
    fprintf(out, "   CAS (->) = %18.15f   abs = %13.5e  rel = %13.5e\n",
                 Sum_ROO, err_1, rel_err_1);


        /***** Double. *****/

    fprintf(out, "\n");
    fprintf(out, " Double:\n");

        /* Double --- iterative sum (forward, backward). */

    sum_1_d = 0.0;
    for (i = 1; i <= n; ++i) {
        clan_d = 1.0 / ( (double) i );
        sum_1_d = sum_1_d + clan_d;
    }
    err_1 = Sum - sum_1_d;
    rel_err_1 = err_1 / Sum;

    sum_2_d = 0.0;
    for (i = n; i >= 1; --i) {
        clan_d = 1.0 / ( (double) i );
        sum_2_d = sum_2_d + clan_d;
    }
    err_2 = Sum - sum_2_d;
    rel_err_2 = err_2 / Sum;

    fprintf(out, "     S (->) = %18.15f   abs = %13.5e  rel = %13.5e\n",
                 sum_1_d, err_1, rel_err_1);
    fprintf(out, "     S (<-) = %18.15f   abs = %13.5e  rel = %13.5e\n",
                 sum_2_d, err_2, rel_err_2);

        /* Double --- Compensated summation without correction. */

    c_sum_1_d = 0.0;
    e_d = 0.0;
    for (i = 1; i <= n; ++i) {
        clan_d = 1.0 / ( (double) i );
        t_d = c_sum_1_d;
        y_d = clan_d + e_d;
        c_sum_1_d = t_d + y_d;
        e_d = ( t_d - c_sum_1_d ) + y_d;
    }
    e_1_d = e_d;
    err_1 = Sum - c_sum_1_d;
    rel_err_1 = err_1 / Sum;

    c_sum_2_d = 0.0;
    e_d = 0.0;
    for (i = n; i >= 1; --i) {
        clan_d = 1.0 / ( (double) i );
        t_d = c_sum_2_d;
        y_d = clan_d + e_d;
        c_sum_2_d = t_d + y_d;
        e_d = ( t_d - c_sum_2_d ) + y_d;
    }
    e_2_d = e_d;
    err_2 = Sum - c_sum_2_d;
    rel_err_2 = err_2 / Sum;

    fprintf(out, "    CS (->) = %18.15f   abs = %13.5e  rel = %13.5e\n",
                 c_sum_1_d, err_1, rel_err_1);
    fprintf(out, "    CS (<-) = %18.15f   abs = %13.5e  rel = %13.5e\n",
                 c_sum_2_d, err_2, rel_err_2);

        /* Double --- Compensated summation + Kahan's correction. */

    k_sum_1_d = c_sum_1_d + e_1_d;
    err_1 = Sum - k_sum_1_d;
    rel_err_1 = err_1 / Sum;

    k_sum_2_d = c_sum_2_d + e_2_d;
    err_2 = Sum - k_sum_2_d;
    rel_err_2 = err_2 / Sum;

    fprintf(out, "    KS (->) = %18.15f   abs = %13.5e  rel = %13.5e\n",
                 k_sum_1_d, err_1, rel_err_1);
    fprintf(out, "    KS (<-) = %18.15f   abs = %13.5e  rel = %13.5e\n",
                 k_sum_2_d, err_2, rel_err_2);


        /***** Float. *****/

    fprintf(out, "\n");
    fprintf(out, " Float:\n");

        /* Float --- iterative sum (forward, backward). */

    sum_1_f = 0.0f;
    for (i = 1; i <= n; ++i) {
        clan_f = 1.0f / ( (float) i );
        sum_1_f = sum_1_f + clan_f;
    }
    err_1 = Sum - sum_1_f;
    rel_err_1 = err_1 / Sum;

    sum_2_f = 0.0f;
    for (i = n; i >= 1; --i) {
        clan_f = 1.0f / ( (float) i );
        sum_2_f = sum_2_f + clan_f;
    }
    err_2 = Sum - sum_2_f;
    rel_err_2 = err_2 / Sum;

    fprintf(out, "     S (->) = %18.15f   abs = %13.5e  rel = %13.5e\n",
                 sum_1_f, err_1, rel_err_1);
    fprintf(out, "     S (<-) = %18.15f   abs = %13.5e  rel = %13.5e\n",
                 sum_2_f, err_2, rel_err_2);

        /* Float --- Compensated summation without correction. */

    c_sum_1_f = 0.0f;
    e_f = 0.0f;
    for (i = 1; i <= n; ++i) {
        clan_f = 1.0f / ( (float) i );
        t_f = c_sum_1_f;
        y_f = clan_f + e_f;
        c_sum_1_f = t_f + y_f;
        e_f = ( t_f - c_sum_1_f ) + y_f;
    }
    e_1_f = e_f;
    err_1 = Sum - c_sum_1_f;
    rel_err_1 = err_1 / Sum;

    c_sum_2_f = 0.0f;
    e_f = 0.0f;
    for (i = n; i >= 1; --i) {
        clan_f = 1.0f / ( (float) i );
        t_f = c_sum_2_f;
        y_f = clan_f + e_f;
        c_sum_2_f = t_f + y_f;
        e_f = ( t_f - c_sum_2_f ) + y_f;
    }
    e_2_f = e_f;
    err_2 = Sum - c_sum_2_f;
    rel_err_2 = err_2 / Sum;

    fprintf(out, "    CS (->) = %18.15f   abs = %13.5e  rel = %13.5e\n",
                 c_sum_1_f, err_1, rel_err_1);
    fprintf(out, "    CS (<-) = %18.15f   abs = %13.5e  rel = %13.5e\n",
                 c_sum_2_f, err_2, rel_err_2);

        /* Float --- Compensated summation + Kahan's correction. */

    k_sum_1_f = c_sum_1_f + e_1_f;
    err_1 = Sum - k_sum_1_f;
    rel_err_1 = err_1 / Sum;

    k_sum_2_f = c_sum_2_f + e_2_f;
    err_2 = Sum - k_sum_2_f;
    rel_err_2 = err_2 / Sum;

    fprintf(out, "    KS (->) = %18.15f   abs = %13.5e  rel = %13.5e\n",
                 k_sum_1_f, err_1, rel_err_1);
    fprintf(out, "    KS (<-) = %18.15f   abs = %13.5e  rel = %13.5e\n",
                 k_sum_2_f, err_2, rel_err_2);

    fclose(out);

    return 0;
}
